#include <stdlib.h>
#include <stdio.h>
#include "libxml.h"

#if defined(LIBXML_THREAD_ENABLED) && defined(LIBXML_CATALOG_ENABLED)
#include <libxml/globals.h>
#include <libxml/threads.h>
#include <libxml/parser.h>
#include <libxml/catalog.h>

#if defined(__WIN32__) && defined(HAVE_WIN32_THREADS)
#include <windows.h>
#define THREAD_TYPE HANDLE
#define THREAD_CREATION_RETURN_TYPE BOOL
#define THREAD_RETURN_CODE_TYPE DWORD
#define THREAD_SPECIFIC_DATA_TYPE DWORD WINAPI
#else
#include <pthread.h>
#define THREAD_TYPE pthread_t
#define THREAD_CREATION_RETURN_TYPE int
#define THREAD_RETURN_CODE_TYPE void *
#define THREAD_SPECIFIC_DATA_TYPE void *
#endif

#include <string.h>
#include <unistd.h>
#include <assert.h>

#define	MAX_ARGC	20
static THREAD_TYPE      tid[MAX_ARGC];

static const char *catalog = "test/threads/complex.xml";
static const char *testfiles[] = {
    "test/threads/abc.xml",
    "test/threads/acb.xml",
    "test/threads/bac.xml",
    "test/threads/bca.xml",
    "test/threads/cab.xml",
    "test/threads/cba.xml",
    "test/threads/invalid.xml",
};

const char *Okay = "OK";
const char *Failed = "Failed";

#ifndef xmlDoValidityCheckingDefaultValue
#error xmlDoValidityCheckingDefaultValue is not a macro
#endif
#ifndef xmlGenericErrorContext
#error xmlGenericErrorContext is not a macro
#endif

static THREAD_SPECIFIC_DATA_TYPE
thread_specific_data(void *private_data)
{
    xmlDocPtr myDoc;
    const char *filename = (const char *) private_data;
    int okay = 1;

    if (!strcmp(filename, "test/threads/invalid.xml")) {
        xmlDoValidityCheckingDefaultValue = 0;
        xmlGenericErrorContext = stdout;
    } else {
        xmlDoValidityCheckingDefaultValue = 1;
        xmlGenericErrorContext = stderr;
    }
    myDoc = xmlParseFile(filename);
    if (myDoc) {
        xmlFreeDoc(myDoc);
    } else {
        printf("parse failed\n");
	okay = 0;
    }
    if (!strcmp(filename, "test/threads/invalid.xml")) {
        if (xmlDoValidityCheckingDefaultValue != 0) {
	    printf("ValidityCheckingDefaultValue override failed\n");
	    okay = 0;
	}
        if (xmlGenericErrorContext != stdout) {
	    printf("xmlGenericErrorContext override failed\n");
	    okay = 0;
	}
    } else {
        if (xmlDoValidityCheckingDefaultValue != 1) {
	    printf("ValidityCheckingDefaultValue override failed\n");
	    okay = 0;
	}
        if (xmlGenericErrorContext != stderr) {
	    printf("xmlGenericErrorContext override failed\n");
	    okay = 0;
	}
    }
    if (okay == 0)
	return((THREAD_RETURN_CODE_TYPE) Failed);
    return ((THREAD_RETURN_CODE_TYPE) Okay);
}

int
main()
{
    unsigned int i, repeat;
    unsigned int num_threads = sizeof(testfiles) / sizeof(testfiles[0]);
    THREAD_RETURN_CODE_TYPE results[MAX_ARGC];
    THREAD_CREATION_RETURN_TYPE ret;

    xmlInitParser();
    for (repeat = 0;repeat < 500;repeat++) {
	xmlLoadCatalog(catalog);

	for (i = 0; i < num_threads; i++) {
	    results[i] = ( THREAD_RETURN_CODE_TYPE ) NULL;
	    tid[i] = (THREAD_TYPE) -1;
	}

	for (i = 0; i < num_threads; i++) {
#if defined(__WIN32__) && defined(HAVE_WIN32_THREADS)
	  DWORD useless;
	  tid[i] = CreateThread (NULL, 0, thread_specific_data, (char*) testfiles[i], 0, &useless);
#else
	  ret = pthread_create(&tid[i], 0, thread_specific_data,
			       (void *) testfiles[i]);
	    if (ret != 0) {
		perror("pthread_create");
		exit(1);
	    }
#endif
	}

#if defined(__WIN32__) && defined(HAVE_WIN32_THREADS)
	if (WaitForMultipleObjects (num_threads, tid, TRUE, INFINITE) == WAIT_FAILED) perror ("WaitForMultipleObjects failed");

	for (i = 0; i < num_threads; i++)
	  {
	    ret = GetExitCodeThread (tid[i], &results[i]);
	    if (ret == 0)
	      {
		perror("GetExitCodeThread");
		exit(1);
	      }
	    CloseHandle (tid[i]);
	  }

#else

	for (i = 0; i < num_threads; i++) {
	    ret = pthread_join(tid[i], &results[i]);
	    if (ret != 0) {
		perror("pthread_join");
		exit(1);
	    }
	}
#endif

	xmlCatalogCleanup();
	for (i = 0; i < num_threads; i++)
	  if (results[i] != (THREAD_RETURN_CODE_TYPE) Okay) {
	    printf("Thread %d handling %s failed\n", i, testfiles[i]);
	  } else {
	    printf("Thread %d handling %s succeeded\n", i, testfiles[i]);
	  }
    }
    xmlCleanupParser();
    xmlMemoryDump();
    return (0);
}

#else /* !LIBXML_THREADS_ENABLED */
int
main(void)
{
    fprintf(stderr, "libxml was not compiled with thread or catalog support\n");
    return (0);
}
#endif
